/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * System Load Average Monitor
 * 
 * Copyright (C) 2005, Digium, Inc.
 *
 * Russell Bryant <russelb@clemson.edu>
 * Kevin P. Fleming <kpfleming@digium.com>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 1.7 $")

#include "asterisk/channel.h"
#include "asterisk/utils.h"
#include "asterisk/config.h"
#include "asterisk/logger.h"
#include "asterisk/sched.h"
#include "asterisk/cli.h"

static struct sched_context *sched;
static int refresh_sched = -1;
static pthread_t refresh_thread = AST_PTHREADT_NULL;
static int refresh_interval;

#define REFRESH_DEFAULT 300     /* refresh every 5 minutes */
#define MAXLOAD_DEFAULT 1.0

double ast_curload = 0;
double ast_maxload = MAXLOAD_DEFAULT;
int option_maxload = 0;
int ast_above_maxload = 0;

static void *do_refresh(void *data)
{
	for (;;) {
		pthread_testcancel();
		usleep(ast_sched_wait(sched));
		pthread_testcancel();
		ast_sched_runq(sched);
	}
	return NULL;
}

static int refresh_load(void *data)
{
	if (getloadavg(&ast_curload, 1) != 1)
		return -1;

	if (ast_curload > ast_maxload)
		ast_above_maxload = 1;
	else
		ast_above_maxload = 0;

	/* automagically reschedule */
	return -1;
}

static int handle_cli_reload(int fd, int argc, char *argv[])
{
	if (argc > 2)
		return RESULT_SHOWUSAGE;

	loadmonitor_reload();

	return 0;
}

static int handle_cli_refresh(int fd, int argc, char *argv[])
{
	if (argc > 2)
		return RESULT_SHOWUSAGE;

	refresh_load(NULL);	

	ast_cli(fd, "Current Load Avg: %lf\n", ast_curload);

	return 0;
}

static int handle_cli_status(int fd, int argc, char *argv[])
{
	if (argc > 2)
		return RESULT_SHOWUSAGE;

	ast_cli(fd, "Load Monitor: %s\n", option_maxload ? "enabled" : "disabled");
	ast_cli(fd, "Refresh Interval: %d seconds\n", refresh_interval);
	ast_cli(fd, "Current Load Avg: %lf\n", ast_curload);
	ast_cli(fd, "Max Load Avg: %lf\n", ast_maxload);
	ast_cli(fd, "Above Threshold: %s\n", ast_above_maxload ? "Yes" : "No");

	return 0;
}

static struct ast_cli_entry cli_reload = {
	.cmda = { "loadmonitor", "reload", NULL },
	.handler = handle_cli_reload,
	.summary = "Reloads the load average monitor configuration",
	.usage = 
	"Usage: loadmonitor reload\n"
	"       Reloads the load average monitor configuration.\n"
};

static struct ast_cli_entry cli_refresh = {
	.cmda = { "loadmonitor", "refresh", NULL },
	.handler = handle_cli_refresh,
	.summary = "Performs an immediate refresh",
	.usage = 
	"Usage: loadmonitor refresh\n"
	"       Peforms an immediate refresh of the system load average.\n",
};

static struct ast_cli_entry cli_status = {
	.cmda = { "loadmonitor", "status", NULL },
	.handler = handle_cli_status,
	.summary = "Display the load average monitor status",
	.usage =
	"Usage: loadmonitor status\n"
	"       Displays the load average monitor status.\n"
};

int loadmonitor_init(void)
{
	int res;

	sched = sched_context_create();
	if (!sched) {
		ast_log(LOG_ERROR, "Unable to create schedule context.\n");
		return -1;
	}

	res = loadmonitor_reload();
	
	if (!res) {
		ast_cli_register(&cli_reload);
		ast_cli_register(&cli_status);
		ast_cli_register(&cli_refresh);
	}	

	refresh_load(NULL);
	
	return res;
}

int loadmonitor_reload(void)
{
	struct ast_config *config;
	const char *interval_value;
	const char *enabled_value;
	const char *maxload_value;
	int interval;
	int was_enabled;
	pthread_attr_t attr;
	int res = -1;

	/* reset defaults in preparation for reading config file */
	refresh_interval = REFRESH_DEFAULT;
	was_enabled = option_maxload;
	option_maxload = 0;

	if (refresh_sched > -1)
		ast_sched_del(sched, refresh_sched);

	if ((config = ast_config_load("loadmonitor.conf"))) {
		if ((enabled_value = ast_variable_retrieve(config, "general", "enable"))) {
			option_maxload = ast_true(enabled_value);
		}
		if ((interval_value = ast_variable_retrieve(config, "general", "refreshinterval"))) {
			if (sscanf(interval_value, "%d", &interval) < 1)
				ast_log(LOG_WARNING, "Unable to convert '%s' to a numeric value.\n", interval_value);
			else if (interval < 0)
				ast_log(LOG_WARNING, "Invalid refresh interval '%d' specified, using default\n", interval);
			else
				refresh_interval = interval;
		}
		if ((maxload_value = ast_variable_retrieve(config, "general", "maxload"))) {
			if ((sscanf(maxload_value, "%lf", &ast_maxload)) != 1)
				ast_log(LOG_WARNING, "Unable to set max load average to '%s'\n", maxload_value);
		}
		ast_config_destroy(config);
	}

	if (option_maxload && refresh_interval) {
		refresh_sched = ast_sched_add(sched, refresh_interval * 1000, refresh_load, NULL);
		ast_log(LOG_NOTICE, "The system load average will be checked every %d seconds.\n", refresh_interval);
	}

	/* if this reload enabled the load monitor, create the background thread
	   if it does not exist */
	if (option_maxload && !was_enabled && (refresh_thread == AST_PTHREADT_NULL)) {
		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
		if (ast_pthread_create(&refresh_thread, &attr, do_refresh, NULL) < 0) {
			ast_log(LOG_ERROR, "Unable to start refresh thread.\n");
			ast_sched_del(sched, refresh_sched);
		}
		else {
			ast_cli_register(&cli_refresh);
			res = 0;
		}
	}
	/* if this reload disabled the load monitor and there is a background thread,
	   kill it */
	else if (!option_maxload && was_enabled && (refresh_thread != AST_PTHREADT_NULL)) {
		/* wake up the thread so it will exit */
		pthread_cancel(refresh_thread);
		pthread_kill(refresh_thread, SIGURG);
		pthread_join(refresh_thread, NULL);
		refresh_thread = AST_PTHREADT_NULL;
		ast_cli_unregister(&cli_refresh);
		res = 0;
	}
	else
		res = 0;

	return res;
}


